
from ..dependencies import pd, dt



def convert_partial_time(partial_time: str) -> dt.datetime.time:
    """
    Convert the partial time in the format '09:02' to a datetime.time object with 0 seconds.

    Parameters:
    ----------
    partial_time : str
        The partial time in the format '09:02'

    Returns:
    -------
    datetime.time
        The time as a `datetime.time` object with 0 seconds.
    """
    try:
        time = dt.datetime.strptime(partial_time, '%H:%M').time()
        time_str = time.strftime('%H:%M:%S')
        time_obj = dt.datetime.strptime(time_str, '%H:%M:%S').time()
    except Exception as e:
        raise ValueError(f"Failed to convert time {partial_time} to datetime object. Error: {e}")

    return time_obj


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Generate averages for each section of the experimental protocol from already ingested finometer data.

This module provides a function to calculate averages for each section of the experimental protocol. 
The resulting averages are concatenated into a single row, allowing for easy merging of participants into a single dataframe. 
Optionally, the function can save the reshaped data as a CSV file.

Notes
-----
- The `frame` parameter should be a pandas DataFrame object containing data produced by `read_raw_finometer_data` function.
- The `id` parameter should be a string representing the participant ID (should also be produced by `read_raw_finometer_data` function).
- The `times` parameter should be a dictionary of tuples representing the time periods for calculating averages. There are functions in the times section to help formatting times.

Example
-------
# Import averaged finometer data
frame, id = read_raw_finometer_data('/path/to/folder')

# Define the time periods for calculating averages
times = {
    'baseline': ('00:03:00, '00:10:00'),
    'task': ('00:10:00', '00:19:00'),
    ...
}

# Generate protocol averages
averages = generate_protocol_averages(frame, id, times=times, save_csv=True)
'''




from ..dependencies import pl, pd, reduce
from .create_chunk import create_chunk


def generate_protocol_averages(frame, id, times=None, save_csv=None):
    '''A function that imports the finometer dataframes (which have already been processed from the raw data)
    to produce averages for each section of the experimental protocol concatenated into a single row so that participants can then be merged into a single dataframe.

    Parameters
    ----------
    frame : pandas.DataFrame 
        The DataFrame containing the averaged finometer data
    id : str
        The participant ID
    save_csv : bool, optional
        If True, the imported data will be saved as a .csv file in the same folder as the .csv file, 
        this is not always needed and should be used sparingly
    times : dict, optional
        A dictionary of tuples of times, with the keys being the names of the time periods.

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the mean values of the given columns during each time period of the study.

    Raises
    ------
    TypeError
        If frame is not a pandas.DataFrame object
        If id is not a string
    ValueError
        If times is not provided as a dictionary with at least one key-value pair
        If there are not enough times provided for a given time period
        If there are too many times provided for a given time period
    '''

    # check if frame is a pandas.DataFrame object
    if not isinstance(frame, pd.DataFrame):
        raise TypeError('''
        frame must be a pandas.DataFrame object, produced by the read_raw_finometer_data function, 
        have you run the read_raw_finometer_data function on the data?''')

    if not isinstance(id, str):
        raise TypeError('id must be a string')

    if not times:
        raise ValueError("times must be a dictionary and at least one key-value pair must be provided.")
    
    # Create an empty list of dataframes, each representing a chunk of the protocol
    chunks = []
    
    for i in times.keys():
        if len(times[i]) < 2:
            raise ValueError(f"There are not enough times provided for the {i}.")
        elif len(times[i]) > 2:
            raise ValueError(f"There are too many times provided for the {i}.")
        elif len(times[i]) == 2:
            if times[i][0] < times[i][1]:
                chunks.append(create_chunk(frame, id, i, times[i][0], times[i][1]))



    data_merge = reduce(lambda left, right: pd.merge(left, right, on=["Participant ID"], how="outer"), chunks)
    data_merge.set_index('Participant ID', inplace=True)

    if save_csv:
        path = pl.Path(save_csv)
        data_merge.to_csv( path / f"{id} protocol_averages.csv")
        print(f"Saved {id} protocol averages.csv to {path.stem}")

    return data_merge


#!/usr/bin/env python
# -*- coding: utf-8 -*-



from ..dependencies import pl, pd, Union, Tuple, Optional, dt

def read_raw_finometer_data(folder_path: Union[str, pl.Path], interval: Optional[str] = None, save_csv: bool = False) -> Tuple[pd.DataFrame, str]:
    """Imports the raw finometer data and calculates the average of each measure over the selected time period.

    This function imports the raw finometer data from the specified folder path or file and calculates the average of each measure over a selected time period. The default time period is 1 minute, but it can be changed by setting the `interval` parameter to a different value. This function is a convenient way to preprocess the data before further analysis.

    Args:
        folder_path (Union[str, pathlib.Path]): The path to the folder containing the .txt file.
        interval (str, optional): If provided, the data is resampled to the given interval. Defaults to None.
        save_csv (bool, optional): If True, the imported data (or the resampled data if `interval` is provided) is saved as a CSV file in the same folder as the data file. Defaults to False.

    Raises:
        TypeError: If `folder_path` is not a pathlib.Path object or a string.
        ValueError: If `folder_path` does not exist or is not a directory.
                    If there is not exactly one .txt file in the folder.

    Returns:
        Tuple[pd.DataFrame, str]: A tuple containing the following:
            - A DataFrame with the raw finometer data resampled to the given interval.
            - The Participant ID of the participant whose data is being imported.

    Notes:
        - The function expects the finometer data to be stored in a single .txt file within the specified folder.
        - The function reads the data from the .txt file, performs preprocessing steps (such as dropping unnecessary columns and converting timestamps), and calculates the average of each measure over the selected time period.
        - If `interval` is provided, the data is resampled to the given interval using the mean value for each resampled interval.
        - If `save_csv` is True, the imported data (or the resampled data if `interval` is provided) is saved as a CSV file in the same folder as the data file.

    Example:
        df, id = read_raw_finometer_data('/path/to/folder', interval='1T', save_csv=True)
    """
    
    try:
        folder_path = pl.Path(folder_path)
    except TypeError:
        raise TypeError('folder_path must be a pathlib.Path object or a string')

    if not folder_path.exists():
        raise ValueError('folder_path does not exist')

    if folder_path.is_dir():
        files = [file for file in folder_path.glob('*.txt')]
        if len(files) != 1:
            raise ValueError(f'Expected one .txt file, but found {len(files)} in the folder')
        file = files[0]
    elif folder_path.is_file():
        file = folder_path

    ID = file.stem.split('_')[0]



    df = pd.read_csv(
        file,
        sep=';',
        header=0,
        skiprows=8,
        skipfooter=1,
        engine='python',
    )

    df = df.drop(df.columns[13], axis=1)

    df['Time (s)'] = pd.to_datetime(df['Time (s)'], format='%H:%M:%S.%f').dt.floor('ms')



    if interval:

        csv_path = folder_path / file.with_stem(f'imported {interval} data for {ID}').with_suffix('.csv')
        try:
            df_resampled = df.set_index(pd.to_datetime(df['Time (s)'], format='%H:%M:%S.%f')).resample(f'{interval}').mean(numeric_only=True)
            df_resampled.index = df_resampled.index.strftime('%H:%M:%S.%f').str[:-3]
        except ValueError:
            raise ValueError(f'{interval} is not a valid time period, valid time periods are: 1s, 1T, 1H, 1D, 1W, 1M, 1Q, 1A')
    else:
        csv_path = folder_path / file.with_stem(f'imported data for {ID}').with_suffix('.csv')
        df = df.set_index(pd.to_datetime(df['Time (s)'], format='%H:%M:%S.%f').dt.strftime('%H:%M:%S.%f').str[:-3])
        df = df.drop('Time (s)', axis=1)


    if save_csv:
        df.to_csv(csv_path, index=True)

    return (df_resampled, ID) if interval else (df, ID)


from ..dependencies import pl,pd

def import_protocol_times(times_file_path: pl.Path or str, add_seconds: bool = False, flatten_seconds: bool = False, save_csv: bool = False) -> pd.DataFrame:
    '''
    This function imports the protocol times from a .csv or excel file file and returns a cleaned pandas dataframe with the protocol times for each participant.

    Parameters
    ----------
    times_file_path : pathlib.Path or str
        The path to the .csv/.xlsx  file containing the protocol times.
    add_seconds : bool, optional
        If True, seconds will be added to the time values (if missing).
    flatten_seconds : bool, optional
        If True, seconds will be set to 00 for all time values.
    save_csv : bool, optional
        If True, the imported data will be saved as a .csv file in the same folder as the .csv file.

    Raises
    ------
    TypeError:
        If times_file_path is not a pathlib.Path object.
    ValueError:
        If times_file_path does not exist or is not a file.
        If times_file_path does not have a .csv extension.
        If add_seconds and flatten_seconds are both True.
        If the function is unable to add seconds to time values or set seconds to 00.

    Returns
    -------
    pandas.DataFrame
        A cleaned pandas dataframe with the protocol times for each participant.
    '''

    def add_seconds_to_time(time_str):
        '''This function adds seconds to the time string for in case the time string is missing seconds'''
        if len(time_str) == 5:
            time_str += ":00"
        return time_str

    def flatten_seconds(time_str):
        '''This function sets seconds to 00 for a given time string'''
        return time_str[:5] + ':00'
    
    supported_extensions = ['.csv', '.xlsx']

    if not isinstance(times_file_path, pl.Path):  # check if file_path is a pathlib.Path object
        raise TypeError('times_file_path must be a pathlib.Path object')
    elif not times_file_path.exists():  # and if it exists
        raise ValueError('times_file_path does not exist')
    elif not times_file_path.is_file():  # and is a file
        raise ValueError('times_file_path is not a file')
    elif times_file_path.suffix not in supported_extensions:  # and is a supported file extension
        raise ValueError('times_file_path is not a supported file (csv or xlsx)')
    else:
        if times_file_path.suffix == '.csv':
            df = pd.read_csv(times_file_path)
        elif times_file_path.suffix == '.xlsx':
            df = pd.read_excel(times_file_path)
        else:
            raise ValueError('Invalid file extension')
        
        if add_seconds and flatten_seconds:
            raise ValueError('Only one of add_seconds and flatten_seconds can be True')

        if add_seconds:
            try:
                df.iloc[:, 1:] = df.iloc[:, 1:].applymap(lambda x: add_seconds_to_time(x) if isinstance(x, str) else x)
                df.iloc[:, 1:] = df.iloc[:, 1:].applymap(lambda x: pd.to_datetime(x, format='%H:%M:%S', errors='coerce'))
            except:
                print('Could not add seconds to time, please check the time format')

        elif flatten_seconds:
            try:
                df.iloc[:, 1:] = df.iloc[:, 1:].applymap(lambda x: add_seconds_to_time(x) if isinstance(x, str) else x)
                df.iloc[:, 1:] = df.iloc[:, 1:].applymap(lambda x: flatten_seconds(x) if isinstance(x, str) else x)
                df.iloc[:, 1:] = df.iloc[:, 1:].applymap(lambda x: pd.to_datetime(x, format='%H:%M:%S', errors='coerce'))
            except:
                print('Could not set seconds to 00, please check the time format')
        
        if save_csv: #if you want to save the csv file (which may be useful if you want to use the data in other ways)
            try:
                df.to_csv(times_file_path.parent / f"cleaned times.csv", index=False)
                print(f"CSV saved for {times_file_path.stem}")
            except Exception as e:
                print(f"Could not save csv file, error: {e}")
            


        return df


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provides a function to read the demographics data (Participant ID, age, height, weight) from the specified folder path and return a DataFrame containing the data.
This can then be used to merge the demographics data with any of the other permutations of the cardiovascular/hemodynamic data produced by these functions.
Notes
-----
- The function expects the finometer data to be stored in a single .txt file within the specified folder.
- The function reads the data from the .txt file, performs preprocessing steps (such as dropping unnecessary columns and converting timestamps), and calculates the average of each measure over the selected time period.


Example
-------
P1_demographics = read_raw_demographics('path/to/folder')
"""

from ..dependencies import pd, pl

def read_raw_demographics(folder_path: str) -> pd.DataFrame:
    """
    Reads in the demographics from the .txt file and returns a DataFrame row containing the data.

    Parameters:
    file_path (str): Path to the demographics file.

    Returns:
    demographics_df (pd.DataFrame): DataFrame containing the demographics data.
    """

    try:
        folder_path = pl.Path(folder_path)
    except TypeError:
        raise TypeError('folder_path must be a pathlib.Path object or a string')

    if not folder_path.exists():
        raise ValueError('folder_path does not exist')

    if folder_path.is_dir():
        files = [file for file in folder_path.glob('*.txt')]
        if len(files) != 1:
            raise ValueError(f'Expected one .txt file, but found {len(files)} in the folder')
        file = files[0]
    elif folder_path.is_file():
        file = folder_path

    ID = file.stem.split('_')[0]

    # Read in the demographics data from the file
    df = pd.read_csv(
        file,
        sep=';',
        header=0,
        skiprows=2,
        nrows=1,
        engine='python'
        )

    # Select the relevant columns from the DataFrame
    demographics_df = df.loc[:, ['Identification', 'Age (yrs)', 'Height (cm)', 'Weight (kg)', 'Gender']]

    # Rename the columns
    demographics_df.columns = ['Participant ID', 'Age (years)', 'Height (cm)', 'Weight (kg)', 'Gender']


    return demographics_df


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
This module provides a function to create a chunk of data from a DataFrame between specified start and end times and return a new DataFrame containing the mean values for each cardiovascular measure during that chunk. 

Notes
-----
 - The DataFrame must have been created by the `read_raw_finometer_data` function. 
 - The time values must be in the format 'HH:MM:SS'. There are functions to help normalise the time values in the fino2py.times module.

Example
-------
baseline_1 = create_chunk(df, 'Participant 1', 'Baseline', '00:00:00', '00:05:00')
'''

from ..dependencies import pd, Union
from ..times.convert_timestamp_time import convert_timestamp_time


def create_chunk(df: pd.DataFrame, ID: str, tag: str, start: Union[str, None], end: Union[str, None]) -> pd.DataFrame:
    """
    Create a chunk of data from a dataframe between specified start and end times and return a new dataframe
    containing the mean values for each cardiovascular measure during that chunk.
    
    Parameters:
    -----------
    df : pandas DataFrame
        The dataframe containing the data to extract a chunk from.
    ID : str
        The participant ID to include in the output dataframe.
    tag : str
        The tag to include in the column names of the output dataframe.
    start : str or None
        The start time of the chunk in the format 'HH:MM:SS' or 'HH:MM:SS.mmm'. If None, the chunk starts at the 
        beginning of the dataframe.
    end : str or None
        The end time of the chunk in the format 'HH:MM:SS' or 'HH:MM:SS.mmm'. If None, the chunk ends at the 
        end of the dataframe.
    
    Returns:
    --------
    pandas DataFrame
        A new dataframe containing the mean values for each column in the specified chunk of the input dataframe.
        The output dataframe has a row for the specified participant ID and columns with names that include the
        specified tag.
    """
    # Convert start and end times to datetime objects if they are specified
    if start:
        try:
            start = convert_timestamp_time(start)
        except:
            raise ValueError(f"Could not convert {start} to datetime object, it must be a string in the format 'HH:MM:SS' or 'HH:MM:SS.mmm'")
    if end:
        try:
            end = convert_timestamp_time(end)
        except:
            raise ValueError(f"Could not convert {end} to datetime object, it must be a string in the format 'HH:MM:SS' or 'HH:MM:SS.mmm'")

    # Extract the chunk of data and compute the mean values for each column
    if start and end:
        chunk = df.loc[start:end].mean().to_frame().T
    elif start:
        chunk = df.loc[start:].mean().to_frame().T
    elif end:
        chunk = df.loc[:end].mean().to_frame().T
    
    # Rename the columns with the specified tag and insert the participant ID as the first column
    chunk.columns = [f"{tag} {i}" for i in chunk.columns]
    chunk.insert(0, 'Participant ID', ID)

    return chunk




from ..dependencies import dt

def convert_timestamp_time(timestamp_time: str) -> str:
    """
    Convert the timestamp string in the format '09:02:12' to '%H:%M:%S' datetime format.

    Parameters:
    ----------
    timestamp_time : str
        The timestamp time in the format '09:02:12'

    Returns:
    -------
    str
        The time in the '%H:%M:%S' format.
    """
    try:
        time = dt.datetime.strptime(timestamp_time, '%H:%M:%S')
    except Exception as e:
        raise ValueError(f"Failed to convert time {timestamp_time} to datetime object. Error: {e}")

    return time.strftime('%H:%M:%S')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provides a function to read the demographics data (Participant ID, age, height, weight) from the specified folder path and return a DataFrame containing the data.
This can then be used to merge the demographics data with any of the other permutations of the cardiovascular/hemodynamic data produced by these functions.
Notes
-----
- The function expects the finometer data to be stored in a single .csv file within the specified folder.
- The function reads the data from the .csv file, performs preprocessing steps (such as dropping unnecessary columns and converting timestamps), and calculates the average of each measure over the selected time period.


Example
-------
P1_demographics = read_raw_demographics('path/to/folder')
"""

from ..dependencies import pd, pl, re, dt, Union, Tuple, Optional, List

def read_nova_demographics(folder_path: str) -> pd.DataFrame:
    """
    Reads in the demographics from the nova press .csv file and returns a DataFrame row containing the data.

    Parameters:
    file_path (str): Path to the demographics file.

    Returns:
    demographics_df (pd.DataFrame): DataFrame containing the demographics data.
    """

    try:
        folder_path = pl.Path(folder_path)
    except TypeError:
        raise TypeError('folder_path must be a pathlib.Path object or a string')

    if not folder_path.exists():
        raise ValueError('folder_path does not exist')

    if folder_path.is_dir(): # If the folder path is a directory, ask the user to specify the file

        files = [file for file in folder_path.glob('*.csv')]
        # print the files in the directory and a number for each file
        print('The following files are in the directory:')
        for i, file in enumerate(files):
            print(f'{i+1}: {file.name}')
        # ask the user to specify the file
        file_number = input('Please enter the number of the file you wish to use: ')
        # check that the input is a number
        if not file_number.isdigit():
            raise ValueError('Input must be a number')
        # check that the input is a valid number
        if int(file_number) not in range(1, len(files)+1):
            raise ValueError(f'Input must be between 1 and {len(files)}')
        # select the file
        file = files[int(file_number)-1]
    elif folder_path.is_file():
        file = folder_path

    ID = file.stem.split(' ')[1]

    detail_row = [4,5]

    # Read in the demographics data from the file
    df = pd.read_csv(
        file,
        sep=';',
        header=0,
        skiprows= lambda x: x not in detail_row,
        engine='python'
        )


    return df


#!/usr/bin/env python
# -*- coding: utf-8 -*-

##Import all functions from the submodules

# The ingesting modules
from .ingesting.read_raw_finometer_data import read_raw_finometer_data
from .ingesting.read_raw_demographics import read_raw_demographics
from .ingesting.read_raw_nova_data import read_raw_nova_data
from .ingesting.read_nova_demographics import read_nova_demographics

# the data cleaning modules
from .cleaning.merge_split_data import merge_split_data


# The reshaping modules
from .reshaping.minute_by_minute import minute_by_minute
from .reshaping.minute_by_minute_from_folder import minute_by_minute_from_folder
from .reshaping.create_chunk import create_chunk
from .reshaping.generate_protocol_averages import generate_protocol_averages
from .reshaping.nova_minute_by_minute import nova_minute_by_minute
from .reshaping.nova_concat import nova_concat

# The times modules
from .times.convert_partial_time import convert_partial_time
from .times.convert_fino_time import convert_fino_time
from .times.convert_timestamp_time import convert_timestamp_time
from .times.import_protocol_times import import_protocol_times


__all__ = [
    'read_raw_finometer_data',
    'read_raw_demographics',
    'read_raw_nova_data',
    'read_nova_demographics',
    'generate_protocol_averages',
    'create_chunk',
    'merge_split_data',
    'minute_by_minute',
    'nova_minute_by_minute',
    'nova_concat',
    'minute_by_minute_from_folder',
    'convert_partial_time',
    'convert_fino_time',
    'convert_timestamp_time',
    'import_protocol_times'
]


#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Append cardiovascular data from a second file to a first file and optionally write the combined data to a new file.
In some cases, due to issue with the finometer during data collection, a single participant may end up with two seperate folders with seperate `.txt` files. 
This module provides a function to append the cardiovascular data from a 'part 2' file to a 'part 1' file. 
The combined data is returned as a list of strings. 
Optionally, the function can write the combined data to an output file. 

Notes
-----
 - This function is intended for use in the data cleaning process, but it should be used cautiously and documented appropriately in the research notes and final report appendices.
 - The function is *not* intended to be used to combine data from two seperate participants.


Example
-------
combined_data = merge_split_data('/path/to/part1_file.txt', '/path/to/part2_file.txt', part2_start_line=9, output_file='/path/to/output_file.txt')

'''


from ..dependencies import pl, Union, Optional, List

def merge_split_data(
    part1_file: Union[str, pl.Path],
    part2_file: Union[str, pl.Path],
    part2_start_line: str = 9, 
    output_file: Optional[Union[str, pl.Path]] = None
) -> List[str]:
    '''
    Appends cardiovascular data from part 2 file to part 1 file and optionally writes the combined data to an output file.

    Parameters
    ----------
    part1_file : Union[str, Path]
        File path of the part 1 file.
    part2_file : Union[str, Path]
        File path of the part 2 file.
    part2_start_line : int, optional
        The line number in the part 2 file from which to start appending the data (default is 9).
        Use this parameter to skip lines in the part 2 file, for example, to account for recalibration.
    output_file : Union[str, Path], optional
        File path of the output file to write the combined data (default is None).

    Returns
    -------
    List[str]
        List of combined cardiovascular data.
    '''
    
    # Step 1: Read in the part 1 file
    with open(part1_file, 'r') as f:
        lines_part1 = f.readlines()

    # Step 2: Save the header above line 9 in file 1
    header_part1 = lines_part1[:8]

    # Step 3: Save the data from line 9 down from file 1
    data_part1 = lines_part1[8:]

    # Step 4: Save the data from line 9 down in file 2
    with open(part2_file, 'r') as f:
        lines_part2 = f.readlines()[part2_start_line:]

    # Step 5: Append the data from file 2 to the bottom of the data from file 1
    combined_data = data_part1 + lines_part2
    
    if output_file is not None:
        # Step 6: Write all three things to a .txt file
        with open(output_file, 'w') as f:
            f.writelines(header_part1)
            f.writelines(combined_data)
    
    return combined_data


#!/usr/bin/env python
# -*- coding: utf-8 -*-


from ..dependencies import pd, Union, Dict 
from .nova_minute_by_minute import nova_minute_by_minute

'''
This function allows us to concatenate the minute-by-minute data for each participant into a single DataFrame so that we can perform group-level analysis.

In future when we have a means of creating protocol 'chunk' averages, we will also be able to use this function to concatenate the chunk averages for each participant into a single DataFrame for group-level analysis. 
But for now this is just for minute-by-minute data.

Notes
-----
- The `data_dict` parameter should be a dictionary with participant IDs as keys and minute-by-minute dataframes as values.
- The `data_dict` should be created using the `read_raw_nova_data` function.
- The Participant ID column is added at the beginning of the DataFrame to allow for sample-level analysis, and to allow for easy merging of various datasets (i.e. demographics data and or survey data from qualtrics or pavlovia).

example
-------
# Import minute-by-minute data for each participant

data_dict: Dict[str, pd.DataFrame] = {}

for sub_folder in project_folder.iterdir():
    try:
        ID, df = read_raw_nova_data(sub_folder, interval='1s', save_csv=True)

        data_dict[ID] = df

    except Exception as e:
        print(f'Error processing folder {sub_folder.name}: {e}')
}

concatenated_df = nova_concat(data_dict)
'''

def nova_concat(data_dict: Dict) -> pd.DataFrame:
    '''
    Concatenate the minute-by-minute data for each participant into a single DataFrame.

    This function takes a dictionary of minute-by-minute dataframes (`data_dict`) and concatenates them into a single DataFrame. The dictionary should be created using the `read_raw_nova_data` function. The Participant ID column is added at the beginning of the DataFrame to allow for sample-level analysis.

    Parameters
    ----------
    data_dict : Dict
        A dictionary containing Participant ID as keys and minute-by-minute dataframes as values.

    Returns
    -------
    pd.DataFrame
        A concatenated DataFrame containing minute-by-minute data for all participants.

    Example
    -------
    data_dict = {
        'Participant1': df1,
        'Participant2': df2,
        ...
    }
    concatenated_df = nova_concat(data_dict)
    '''

    frames = [] 

    for df_id, df in data_dict.items():
        try:
            df = nova_minute_by_minute(df, df_id)
            frames.append(df)
        except Exception as e:
            print(f'Error processing folder {df_id}: {e}')

    # Find the widest dataframe
    widest = max(frames, key=lambda x: x.shape[1])

    # Add the missing columns to each dataframe
    frames = [df.reindex(widest.columns, axis=1) for df in frames]

    # Concatenate the dataframes on the Participant ID and Time(sec) columns
    df = pd.concat(frames)

    return df


#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
This module provides a function for reshaping participant's minute-by-minute data into a single row.

The main function, `nova_minute_by_minute`, takes a DataFrame containing Novapress data and performs reshaping of the data. 
It sets the index as the "Time (s)" column to preserve the original time information. 
The DataFrame is then flattened, creating a new DataFrame with each minute's data in separate columns. The column names are generated by appending "_minute_i" to the original column names, where "i" represents the minute index. The Participant ID is added as a column at the beginning of the new DataFrame, facilitating concatenation of participant data for sample-level analysis.

Note: The `minute_by_minute` function assumes that the input DataFrame contains the necessary columns and structure for reshaping. Also, as the NovaPress data is spread across multiple files, this function is all that's required to reshape the data.

Functions
---------
minute_by_minute(df: pd.DataFrame, df_id: str) -> pd.DataFrame:
    Reshape the participant's minute-by-minute data into a single row.
    Returns a DataFrame with each minute's data in separate columns.

    Parameters
    ----------
    df : pandas DataFrame
        The DataFrame containing finometer data.
    df_id : str
        The Participant ID associated with the data.

    Returns
    -------
    pandas DataFrame
        The reshaped minute-by-minute data DataFrame.
        Each minute's data is in separate columns, with column names reflecting the original column names and minute indices. The Participant ID column is included at the beginning.
'''


from ..dependencies import pd, Tuple #importing pandas and Tuple from dependencies.py

def nova_minute_by_minute(df: pd.DataFrame, df_id: str) -> pd.DataFrame:
    '''
    Reshape the participant's minute by minute data into a single row 

    This function takes a DataFrame containing finometer data (`df`). It converts the index of the DataFrame into a column named "Time (s)" to preserve the original time information.

    The DataFrame is then flattened to create a new DataFrame (`flat_frame`) with each minute's data in a separate column. The column names are generated by appending "_minute_i" to the original column names, where "i" represents the minute index.

    The Participant ID (`df_id`) is added as a column at the beginning of the `flat_frame` DataFrame. This allows for participant data to be easily concatenated into a single DataFrame for sample level analysis.

    Parameters
    ----------
    df : pandas DataFrame
        The DataFrame containing the finometer data.
    df_id : str
        The Participant ID associated with the data.

    Returns
    -------
    pandas DataFrame
        The minute-by-minute analysis DataFrame.

        The `flat_frame` DataFrame contains each minute's data in separate columns, with column names reflecting the original column names and minute indices. The Participant ID column is included at the beginning.
    '''

    df['Time(sec))'] = df.index
    flat_array = df.values.flatten()
    flat_frame = pd.DataFrame(flat_array).T




    new_columns = [[f"{col}_minute_{i}" for col in df.columns] for i in range(df.shape[0])]
    new_columns_flat = [item for sublist in new_columns for item in sublist]



    flat_frame.columns = new_columns_flat
    flat_frame.insert(0, 'Participant ID', df_id)
   

    return flat_frame


#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""

This module provides a function to read the raw NovaPress data from the specified folder path or file and return a DataFrame containing the data.

This function imports the raw NovaPress data from the specified folder path or file and calculates the average of each measure over a selected time period. 
The default time period is 1 secong, but it can be changed by setting the `interval` parameter to a different value. 
This function is a convenient way to preprocess the data before further analysis. 
By default it combines data from the 'Basic Nova' and 'Advanced_Hemodynamics' CSV files by matching on the 'Time(sec)' column. 
This can be changed by setting the `required_files` parameter to a different list of strings specifying the names of the CSV files to be imported and processed.

Notes
-----
- The function expects the data to be stored in one or more .csv files in the specified folder.
- The function reads the data from the CSV files, performs preprocessing steps (such as dropping unnecessary columns and converting timestamps), and calculates the average of each measure over the selected time period.
- The date is extracted from the file name and used to convert the 'Time(sec)' column to a datetime object.
- If `interval` is provided, the data is resampled to the given interval using the mean value for each resampled interval, however the default interval is 1 
second because otherwise we end up with duplicate timestamps and a lot of NaN values.

Example
-------
df, id = read_raw_nova_data('/path/to/folder', interval='1T', save_csv=True)

"""

from ..dependencies import pd, pl, re, dt, Union, Tuple, Optional, List, reduce



def read_raw_nova_data(
    folder_path: Union[str, pl.Path],
    interval: Optional[str] = '1s',
    save_csv: bool = False,
    required_files: list = ['Basic Nova', 'Advanced_Hemodynamics'],
) -> Tuple[str, pd.DataFrame]:
    """
    Imports the raw NovaPress data, processes it, and calculates the average of each measure over the selected time period.

    This function imports the raw NovaPress data from the specified folder path or file and calculates the average of each measure over a selected time period. The default time period is 1 minute, but it can be changed by setting the `interval` parameter to a different value. This function is a convenient way to preprocess the data before further analysis. It combines data from the 'Basic Nova' and 'Advanced_Hemodynamics' CSV files by matching on the 'Time(sec)' column.

    Args:
        folder_path (Union[str, pathlib.Path]): The path to the folder containing the .csv files.
        file_prefix (str, optional): If provided, only files with names starting with this prefix will be considered. Defaults to None.
        interval (str, optional): The interval over which the average of each measure is calculated. Defaults to '1s'.
        save_csv (bool, optional): If True, the imported data (or the resampled data if `interval` is provided) is saved as a CSV file in the same folder as the data file. Defaults to False.
        required_files (list, optional): A list of strings specifying the names of the CSV files to be imported and processed. Defaults to ['Basic Nova', 'Advanced_Hemodynamics'].

    Raises:
        TypeError: If `folder_path` is not a pathlib.Path object or a string.
        ValueError: If `folder_path` does not exist or is not a directory, or if no matching CSV files are found.

    Returns:
        Tuple[str, pd.DataFrame]: A tuple containing the following:
            - A string representing the Participant ID of the participant whose data is being imported.
            - A pandas DataFrame with the raw NovaPress data resampled to the given interval (if provided).

    Notes:
        - The function expects the data to be stored in one or more .csv files in the specified folder.
        - If `file_prefix` is provided, only files with names starting with the specified prefix will be considered.
        - The function reads the data from the CSV files, performs preprocessing steps (such as dropping unnecessary columns and converting timestamps), and calculates the average of each measure over the selected time period.
        - If `interval` is provided, the data is resampled to the given interval using the mean value for each resampled interval.
        - If `save_csv` is True, the imported data (or the resampled data if `interval` is provided) is saved as a CSV file in the same folder as the data file.

    Example:
        df, id = read_raw_nova_data('/path/to/folder', interval='1T', save_csv=True)
    """
    
    try:
        folder_path = pl.Path(folder_path)
    except TypeError:
        raise TypeError('folder_path must be a pathlib.Path object or a string')

    if not folder_path.exists() or not folder_path.is_dir():
        raise ValueError('folder_path does not exist or is not a directory')

    files = []
    for required_file in required_files:
        matching_files = [file for file in folder_path.glob(f'*{required_file}*.csv')]
        files.extend(matching_files)

    if not files:
        raise ValueError('No matching CSV files found in the folder for the specified file prefixes')

    dfs = []

    for file in files:
        # Extract the date string from the file name
        match = re.search(r'(\d{2}-[A-Za-z]{3}-\d{2})', file.name)
        if match:
            date_str = match.group(1)
        else:
            raise ValueError(f"Date pattern not found in the file name: {file.name}")

        # Convert the date string to a datetime object
        file_date = dt.datetime.strptime(date_str, '%d-%b-%y')

        # Read the CSV file
        df = pd.read_csv(file, sep=';', skiprows=7)

        # Convert 'Time(sec)' column to datetime using the extracted date
        df['Time(sec)'] = file_date + pd.to_timedelta(df['Time(sec)'] * 1000, unit='ms')

        # resample the data to the specified interval
        if interval:
            df = df.resample(interval, on='Time(sec)').mean(numeric_only=True)
        
        dfs.append(df)

    # Merge all dataframes on the specified columns
    combined_df = reduce(lambda left, right: pd.merge(left, right, on=['Time(sec)', 'Region'], how='outer'), dfs)
    # drop any 'Unnamed: ' column
    combined_df = combined_df.loc[:, ~combined_df.columns.str.contains('^Unnamed')]

    ID = folder_path.name.split(' ')[1]

    if save_csv:
        combined_df.to_csv(folder_path / f'{ID} imported {interval} data.csv', index=True)

    return ID, combined_df


#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Reshape finometer data into a minute-by-minute format from files in a folder.

This module provides a function to read the raw finometer data from the specified folder path and reshape the data into a minute-by-minute format. 
It utilizes the `read_raw_finometer_data` function to import the data and calculate the average of each measure over the selected time interval (default is 1 minute). 
The `minute_by_minute` function is then applied to reshape the data into a single row, with each column representing the average of a particular measure over each minute of the data collection.

Notes
-----
- The `read_raw_finometer_data` function is used to import the raw finometer data and calculate the average over the selected time interval.
- The `minute_by_minute` function is applied to reshape the data into a single row for each minute.
- If `save_raw` is True, the imported raw data is saved as a CSV file in the same folder as the data files.
- If `save` is True, the minute-by-minute reshaped data is saved as a CSV file in the same folder as the data files.

Example
-------
frame, id = minute_by_minute_from_folder('/path/to/folder', int='1T', save_raw=True, save=True)
'''


from ..dependencies import pd, Tuple
from ..ingesting.read_raw_finometer_data import read_raw_finometer_data
from .minute_by_minute import minute_by_minute

def minute_by_minute_from_folder(path: str, save_raw: bool = False, save: bool = False) -> Tuple[pd.DataFrame, str]:
    '''

    This function reads the raw finometer data from the specified folder path and 'reshapes' the data into a minute-by-minute format. 
    It uses the `read_raw_finometer_data` function to import the data and calculate the average of each measure over the selected time interval (1 minute in this case). 
    Then, the `minute_by_minute` function is applied to reshape the data into a single row, with each column representing the average of a particular measure over each minute of the data collection.

    Parameters
    ----------
    path : str
        The path to the folder containing the finometer data files.
    save_raw : bool, optional
        Specifies whether to save the imported raw data as a CSV file. The default is False.
    save : bool, optional
        Specifies whether to save the minute-by-minute analysis result as a CSV file. The default is False.

    Returns
    -------
    Tuple[pd.DataFrame, str]
        A tuple containing the minute-by-minute analysis DataFrame and the Participant ID associated with the data.

    Notes
    -----
    - The `read_raw_finometer_data` function is used to import the raw finometer data and calculate the average over the selected time interval.
    - The `minute_by_minute` function is applied to reshape the data into a single row for each minute.
    - If `save_raw` is True, the imported raw data is saved as a CSV file in the same folder as the data files.
    - If `save` is True, the minute-by-minute analysis result is saved as a CSV file in the same folder as the data files.

    Example
    -------
    frame, id = minute_by_minute_from_folder('/path/to/folder', int='1T', save_raw=True, save=True)
    '''

    frame, id = read_raw_finometer_data(path, interval= '1T', save_csv=save_raw)

    minute_frame = minute_by_minute(frame, id)
    
    if save == True:
        minute_frame.to_csv(path / f'{id} minute_by_minute.csv', index = False)

    return minute_frame, id


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
This script is just housing the dependencies for the other scripts in the src folder.
It is not intended to be run as a script.
The dependencies are:
    - pathlib
    - pandas
possible future dependencies:
    - csv
'''

import pandas as pd
import pathlib as pl
from functools import reduce
from typing import Union, Tuple, Optional, List, Dict
import datetime as dt
import re



from ..dependencies import dt

def convert_fino_time(fino_time: str) -> dt.datetime.time:
    """
    Converts the string times produced by the Finometer to datime objects in the '%H:%M:%S.%f' format.

    Parameters:
    ----------
    fino_time : str
        The Finometer time in the format '%H:%M:%S.%f'

    Returns:
    -------
    datetime.time
        The time as a `datetime.time` object.
    """
    try:
        time = dt.datetime.strptime(fino_time, '%H:%M:%S.%f').time()
        time_str_no_ms = time.strftime('%H:%M:%S')
        time_obj_no_ms = dt.datetime.strptime(time_str_no_ms, '%H:%M:%S').time()
    except Exception as e:
        raise ValueError(f"Failed to convert time {fino_time} to datetime object. Error: {e}")

    return time_obj_no_ms